----- BEGIN copypasta

local pairs,set in _ENV

local function tech_type(type) 
  return function(techs)
    for k,v in pairs(techs)
      v.type=type

      if not v.cost
        warning('missing cost for '..k)
      end

      v.cost=v.cost or 0

      TECHNOLOGY[k]=v
    end
  end
end



local function binner(t)
  local mult=#t     
  return function(v)
    v*=mult
    for k,i in spairs(t)
      if v<=k 
        return i 
      end
    end
    return t[mult]
  end

end

local mark = mark_table_declaration

--!! COST definitions

local early = binner 
  { 60, 80, 100, 120, 150, 200, 250, 300, 360, 420, 480  }
--  0   0.1 0.2  0.3  0.4  0.5  0.6  0.7  0.8  0.9   1


local mid = binner
--jwf  { 500, 600, 700, 800, 900, 1000, 1200, 1500, 2000, 2500, 3000}
  { 500, 600, 700, 800, 900, 1000, 1500, 2000, 3000, 4000, 5000}
--  0    0.1  0.2  0.3  0.4  0.5    0.6  0.7    0.8  0.9   1

local late = binner
--jwf  {3000, 3500, 4000, 5000, 6000, 7000, 8000, 10000, 15000, 18000, 20000 }
  {6000, 7000, 8000, 9000, 10000, 12500, 15000, 20000, 25000, 30000, 40000 }
--  0     0.1  0.2   0.3   0.4    0.5  0.6   0.7    0.8   0.9  1


local convert_deps = convert_to_set ('deps',true)
local convert_unlock_hulls = convert_to_set('required_by_hull',false)
local convert_parts = convert_to_set('parts',false)
local convert_required_by_part = convert_to_set('required_by_part',false)

local known_fields = set(
  'effect','desc',
  'cost','locked',
  'prereq',
  'race','field','initial',
  'declared',
  'set_base_drive_speed','set_base_drive_range','set_base_scan_range','set_base_command_range',
	'inc_bonus_drive_speed','inc_bonus_drive_range',
  'tactical_speed_boost',
	'specials',
  'advanced_mods',
  'unlocks_designs','unlocks_parts','unlocks_mods','unlocks_ground_units',
  
  'inc_defenses',
  'unlocks_improvement','unlocks_activity',

  'misc_on_hovers',

	'herald_codex','shortened_name','min_herald_attitude',

  'set_trade_per_market',
  'set_lab_level','set_factory_level','set_defense_level',
  'set_mine_level','set_farm_level','set_market_level',
  'set_trade_capacity_mult',

  'inc_improvement_slots',
  'inc_pop_growth',
  'inc_pop','inc_orbitals','inc_global_ground_bonus',

  'set_armor_level', 'set_militia_level', 'set_bunker_level',
  'set_ground_shield_level','set_tank_armor_level',

  )

local function check_fields(t,dinfo)
  
  if dinfo==nil
    dinfo=t
  end



  for k in pairs(t)
    if not known_fields[k]
      error  | format('unknown field %q\n /%s:%s', tostring | k,tostring | dinfo.declared.file, tostring | dinfo.declared.line ) 
    end

    for r,overrides ; t.race?

      check_fields(overrides,t)
     
    end

  end
end



local function inc_desc(t,name)
  local rval

  local function write(t)
    if rval
      rval..=' '..t
    else
      rval=t
    end
  end


  if t.inc_orbitals
    if t.inc_orbitals==1
      write | 'All planets can support one additional orbital structure.'
    else
      warning('missing inc test for', name)
    end
  end

  return rval

end

local function tech(name)
  
  return function(t)

    local _TRACE_INFO = name

    mark(t,1)

    if t.desc 
      t.desc = desc_format(t.desc)
    end
    if t.effect
      t.effect = desc_format(t.effect)
    else
      t.effect = inc_desc(t,name)
    end

    if not t.prereq
      t.prereq=false
    elseif type(t.prereq)=='table'
      if not (t.prereq.type == 'all_of' or t.prereq.type == 'one_of')
        error "unrecognized prereq type"
      end
    end

    LONG_INFO[name] = t.desc or t.effect
    if t.desc and t.effect
      LONG_INFO[name] = format('%s\n\nEffects: %s',t.desc,t.effect) 
    end

    check_fields(t)

    return catch(name,t)
  end
end

local function typelist (field)

  return function(t)

    for i,v in ipairs(t)
      local tname, info = v()
      info.field=field

      if not info.cost
        warning(format('missing cost for %s\n /%s:%d',tname,info.declared.file,info.declared.line ) )
      end

      TECHNOLOGY[tname!] = info!
    end

  end

end

local function inc_planet(restrictions,inc)

  inc = inc or 1

  local match=extract_abbrev_sets(restrictions)

  return function(planet)
    
    for k,ok in pairs(match)
      if not ok[planet[k]]
        return 0
      end
    end

    return inc
  end
end

local function race_specific(races)
  return function (t)
    t.locked=true
    t.race=t.race or {}
    for _, r in ivalues | races
      t.race[r]=t.race[r] or {}
      
      t.race[r].locked=false
    end
    return t
  end
end

function one_of(...)
  return {
    type = 'one_of',
    ...
  }
end

function all_of(...)
  return {
    type = 'all_of',
    ...
  }
end

----- END copypasta


typelist 'Missiles' {

  tech 'Fusion Anti-Missiles' {
    prereq = all_of('Fusion Warheads', 'Anti-Missiles'),
    cost = 0,
    unlocks_parts = 'Fusion Anti-Missiles',
    desc = [=[
    Now that we have reactors that can create and sustain a
    controlled fusion reaction, we should be able to design missiles that
    utilize a miniature version of this reactor, which can both power the
    missile's propulsion and double as its warhead.
     ]=],
  },

  tech 'Antimatter Anti-Missiles' {
    prereq = all_of('Antimatter Warheads', 'Anti-Missiles'),
    cost = 0,
    unlocks_parts = 'AM Anti-Missiles',
  },

  --need to override the old techs so they don't provide anti-missiles by themselves anymore
  tech 'Fusion Warheads'  {
    prereq = all_of('Guidance Systems','Fusion'),
    cost = mid(.2),
    unlocks_parts = 'Fusion Missiles',
    desc = [=[
    Now that we have reactors that can create and sustain a
    controlled fusion reaction, we should be able to design missiles that
    utilize a miniature version of this reactor, which can both power the
    missile's propulsion and double as its warhead.
     ]=],
  },

  tech 'Antimatter Warheads' {
    prereq = all_of('Fusion Warheads','Antimatter'),
    cost = mid(.8),
    unlocks_parts = 'Antimatter Missiles',
  },

}

-- fix for a bizzare vanilla bug that appears to be the result of 
-- leftovers of a once-existing AM anti-missile technology
_G.allowed_aliases['AM Anti-Missiles'] = 'Antimatter Anti-Missiles'